/*
 * uevent.c -- reads uevents from the kernel
 *
 * Copyright 2006-2008 Red Hat, Inc.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author(s): Peter Jones <pjones@redhat.com>
 */

#define _GNU_SOURCE 1

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/prctl.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <termios.h>

#include <argz.h>
#include <envz.h>

#include <nash.h>
#include "util.h"
#include "lib.h"

#include <asm/types.h>
#include <linux/netlink.h>

#ifndef NETLINK_KOBJECT_UEVENT
#define NETLINK_KOBJECT_UEVENT 15
#endif

#ifndef PR_SET_NAME
#define PR_SET_NAME 15
#endif

struct nash_uevent_handler {
    nashContext *nc;
    int socket; 

    char *msg;
    char *path;
    char **envp;
    char *envz;
};

struct nash_uevent_handler *
nashUEventHandlerNew(nashContext *nc)
{
    struct nash_uevent_handler *handler = NULL;
    struct sockaddr_nl sa;

    if ((handler = calloc(1, sizeof (*handler))) == NULL) {
        nashLogger(nc, NASH_ERROR, "could not allocate uevent handler: %m\n");
        return NULL;
    }
    handler->nc = nc;

    handler->socket = socket(PF_NETLINK, SOCK_DGRAM, NETLINK_KOBJECT_UEVENT);
    if (handler->socket < 0) {
        nashLogger(nc, NASH_ERROR, "could not open netlink socket: %m\n");
        free(handler);
        return NULL;
    }
    setFdCoe(handler->socket, 1);

    memset(&sa, '\0', sizeof (sa));
    sa.nl_family = AF_NETLINK;
    sa.nl_pid = getpid();
    sa.nl_groups = -1;

    if (bind(handler->socket, (struct sockaddr *)&sa, sizeof (sa)) < 0) {
        nashLogger(nc, NASH_ERROR, "could not bind to netlink socket: %m\n");
        close(handler->socket);
        free(handler);
        return NULL;
    }
    return handler;
}

int
nashUEventHandlerGetFd(struct nash_uevent_handler *handler)
{
    return handler->socket;
}

void _nashUEventHandlerDestroy(struct nash_uevent_handler **handler)
{
    close((*handler)->socket);
    free(*handler);
    *handler = NULL;
}

static int
get_netlink_msg(int fd, nashUEvent *uevent)
{
    size_t len;
    ssize_t size;
    static char buffer[2560];
    char *pos;
    char *msg = NULL, *path = NULL, *envz = NULL;
    char *argv[] = { NULL };
    size_t envz_len;
    error_t errnum;

    size = recv(fd, &buffer, sizeof (buffer), 0);
    if (size < 0)
        return -1;

    if ((size_t)size > sizeof (buffer) - 1)
        size = sizeof (buffer) - 1;
    buffer[size] = '\0';

    len = strcspn(buffer, "@");
    if (!buffer[len])
        return -1;

    if ((errnum = argz_create(argv, &envz, &envz_len)) > 0)
        goto err;

    pos = buffer;
    msg = strndup(pos, len++);
    pos += len;
    path = strdup(pos);

    pos += strlen(pos) + 1;
    if (len < size + 1) {
        while (pos[0]) {
            char *value = strchr(pos, '=');
            if (value)
                *(value++) = '\0';

            if ((errnum = envz_add(&envz, &envz_len, pos, value)) > 0)
                goto err;
            pos += strlen(pos) + 1;
            if (*pos)
                pos += strlen(pos) + 1;
        }
    }

    uevent->msg = msg;
    uevent->path = path;
    uevent->envz = envz;
    uevent->envz_len = envz_len;
    return 0;
err:
    if (msg)
        free(msg);
    if (path)
        free(path);
    while(envz)
        argz_delete(&envz, &envz_len, envz);
    errno = errnum;
    return -1;
}

/*
 * Minor dragon here -- success is return of >= 0; timeout is return of
 * 0 and speczero(timeout).
 */
int 
nashGetUEventPoll(nashContext *nc, struct timespec *timeout,
    nashUEvent *uevent, struct pollfd *inpd, int npfds)
{
    struct nash_uevent_handler *handler = nc->uh;
    struct pollfd pds[npfds+1];
    struct timespec to, *top = &to;
    int errnum;
    int rc;

    memset(uevent, '\0', sizeof (*uevent));
    memcpy(pds, inpd, npfds * sizeof (*pds));
    pds[npfds].events = POLLIN | POLLPRI | POLLERR | POLLHUP | POLLMSG;
    pds[npfds].revents = 0;
    pds[npfds].fd = handler->socket;

    speccpy(top, timeout, 1, -1);
    do {
        rc = nash_ppoll(pds, npfds+1, top, NULL, 0);
        speccpy(timeout, top, 0, -1);
    } while (rc < 0 && errno == EINTR);
    errnum = errno;

    if (rc > 0) {
        int i;
        for (i = 0; i <= npfds; i++) {
            if (pds[i].fd == handler->socket && pds[i].revents) {
                if (get_netlink_msg(handler->socket, uevent) < 0) {
                    nashLogger(nc, NASH_ERROR, "get_netlink_msg returned %m\n");
                }
                rc--;
            }
        }
        if (npfds)
            memmove(inpd, pds, npfds * sizeof (*pds));
    }

    errno = errnum;
    return rc;
}

/*
 * vim:ts=8:sw=4:sts=4:et
 */
